Require Import Terms.
Require Import LNaVSyntax.
Require Import LNaVBigStep. (* reusing helpers *)

(** * Small-step Semantics *)

Inductive Frame : Type :=
| RLet : Var -> Tm -> Frame
| RRet : Env -> Frame
| RLoc : Frame
| RBrk : Lab -> Lab -> Frame. (* final label + original pc *)

Definition Cont := list Frame.

Inductive RCfg : Type :=
| CR : Tm -> RCfg    (* term to be reduced *)
| CA : (prod Box Lab) -> RCfg.  (* return atom *)

Definition Cfg := prod (prod (prod (prod Lab Env) Atom (* trace *)) Cont) RCfg.

Notation "<< pc , r , s , k , X >>" := ((((pc, r), s), k), X) (at level 5).

(** The reduction relation. *)
Reserved Notation "c1 --> c2" (at level 80).
Inductive step : Cfg -> Cfg -> Prop :=
  (* var *)
  | step_var : forall pc r s k x a,
    maps r x a ->
    <<pc, r, s, k, CR (TVar x)>> --> <<pc, r, s, k, CA a>>
  (* const *)
  | step_const : forall pc r s k c,
    <<pc, r, s, k, CR (TConst c)>> --> <<pc, r, s, k, CA (V (VConst c),bot)>>
  (* let *)
  | step_let_start : forall pc r s k x t1 t2,
    << pc, r, s, k, CR (TLet x t1 t2) >> --> << pc, r, s, RLet x t2 :: k, CR t1 >>
  | step_let_bind : forall pc r s k x t a,
    << pc, r, s, RLet x t :: k, CA a >> --> << pc, (x,a) :: r, s, RRet r :: k, CR t >>
  (* abs *)
  | step_abs : forall pc r s k x t,
    << pc, r, s, k, CR (TAbs x t) >> --> 
    << pc, r, s, k, CA (V (VClos r x t),bot) >>
  (* app *)
  | step_app : forall pc r s k x1 x2 r' x t l a2,
    maps r x1 (V (VClos r' x t),l) ->
    maps r x2 a2 ->
    << pc, r, s, k, CR (TApp x1 x2) >> --> 
    << pc\_/l, (x,a2) :: r', s, RRet r :: k, CR t >>
  | step_app_error : forall pc r s k x1 x2 b l,
    maps r x1 (b,l) ->
    ~ is_box is_abs b ->
    << pc, r, s, k, CR (TApp x1 x2) >> --> 
    << pc\_/l, r, s, k, CA (propagate_nav b s,bot) >>
  (* return *)
  | step_ret : forall pc r s k r' a,
    << pc, r, s, RRet r' :: k, CA a >> --> << pc, r', s, k, CA a >>
  (* prod *)
  | step_prod : forall pc r s k x1 x2 a1 a2,
    maps r x1 a1 ->
    maps r x2 a2 ->
    << pc, r, s, k, CR (TProd x1 x2) >> --> << pc, r, s, k, CA (V (VProd a1 a2),bot) >>
  (* proj *)
  | step_proj : forall pc r s k d x a1 a2 l,
    maps r x (V (VProd a1 a2),l) ->
    << pc, r, s, k, CR (TProj d x) >> --> << pc\_/l, r, s, k, CA (d_choose d a1 a2) >>
  | step_proj_error : forall pc r s k d x b l,
    maps r x (b,l) ->
    ~ is_box is_prod b ->
    << pc, r, s, k, CR (TProj d x) >> --> << pc\_/l, r, s, k, CA (propagate_nav b s,bot) >>
  (* inx *)
  | step_inx : forall pc r s k d x a,
    maps r x a ->
    << pc, r, s, k, CR (TInx d x) >> --> << pc, r, s, k, CA (V (VInx d a),bot) >>
  (* match *)
  | step_match : forall pc r s k x y t1 t2 d a l,
    maps r x (V (VInx d a),l) ->
    << pc, r, s, k, CR (TMatch x y t1 t2) >> -->
    << pc\_/l, (y,a) :: r, s, RRet r :: k, CR (d_choose d t1 t2) >>
  | step_match_error : forall pc r s k x y t1 t2 b l,
    maps r x (b,l) ->
    ~ is_box is_sum b ->
    << pc, r, s, k, CR (TMatch x y t1 t2) >> -->
    << pc\_/l, r, s, k, CA (propagate_nav b s,bot) >>
  (* tag *)
  | step_tag : forall pc r s k x b l,
    maps r x (b,l) ->
    << pc, r, s, k, CR (TTag x) >> --> << pc, r, s, k, CA (tag_box b,l) >>
  (* bop *)
  | step_bop : forall pc r s k bo x1 x2 b1 l1 b2 l2,
    maps r x1 (b1,l1) ->
    maps r x2 (b2,l2) ->
    << pc, r, s, k, CR (TBOp bo x1 x2) >> --> 
    << pc, r, s, k, CA (bop_box bo b1 b2 s, l1 \_/ l2) >>
  (* brackets *)
  | step_bracket_start : forall pc r s k x t l l',
    maps r x (V (VConst (CLab l)),l') ->
    << pc, r, s, k, CR (TBracket x t) >> --> 
    << pc\_/l', r, s, RBrk l (pc\_/l') :: k, CR t >>
  | step_bracket_start_error : forall pc r s k x t b l',
    maps r x (b,l') ->
    ~ is_box is_lab b ->
    << pc, r, s, k, CR (TBracket x t) >> --> 
    << pc\_/l', r, s, k, CA (propagate_nav b s,bot) >>
  | step_bracket_end : forall pc r s k l' pc' b l'',
    << pc, r, s, RBrk l' pc' :: k, CA (b,l'') >> -->
    << pc', r, s, k, CA ((if flows_dec (l'' \_/ pc) (l' \_/ pc') then b else D eBracket s empty),l') >>
  (* labelof *)
  | step_labelof : forall pc r s k x b l,
    maps r x (b,l) ->
    << pc, r, s, k, CR (TLabelOf x) >> --> << pc, r, s, k, CA (V (VConst (CLab l)),bot) >>
  (* get_pc *)
  | step_getpc : forall pc r s k,
    << pc, r, s, k, CR TGetPc >> --> << pc, r, s, k, CA (V (VConst (CLab pc)),bot) >>
  (* mk_nav *)
  | step_mknav : forall pc r s k x b l, 
    maps r x (b,l) ->
    << pc, r, s, k, CR (TMkNav x) >> --> << pc, r, s, k, CA (mk_nav_box b s,l) >>
  (* to_sum *)
  | step_tosum : forall pc r s k x b l,
    maps r x (b,l) ->
    << pc, r, s, k, CR (TToSum x) >> --> << pc, r, s, k, CA (to_sum_box b,l) >>
  (* loc *)
  | step_loc : forall pc r s k loc t,
    << pc, r, s, k, CR (TLoc loc t) >> --> << pc, r, push loc bot s, RLoc :: k, CR t >>
  | step_locret : forall pc r loc s k b l,
    << pc, r, 
       (V (VInx DRight (V (VProd (V (VConst (CLoc loc)),bot) s),bot (* bot or pc'?*))),bot),
       RLoc :: k, CA (b,l) >> --> << pc, r, s, k, CA (trace_propagation b loc bot,l) >>
where "c1 --> c2" := (step c1 c2).
Hint Constructors step.

Tactic Notation "step_cases" tactic(first) ident(c) :=
  first; [
    Case_aux c "step_var"
  | Case_aux c "step_const"
  | Case_aux c "step_let_start"
  | Case_aux c "step_let_bind"
  | Case_aux c "step_abs"
  | Case_aux c "step_app"
  | Case_aux c "step_app_error"
  | Case_aux c "step_ret"
  | Case_aux c "step_prod"
  | Case_aux c "step_proj"
  | Case_aux c "step_proj_error"
  | Case_aux c "step_inx"
  | Case_aux c "step_match"
  | Case_aux c "step_match_error"
  | Case_aux c "step_tag"
  | Case_aux c "step_bop"
  | Case_aux c "step_bracket_start"
  | Case_aux c "step_bracket_start_error"
  | Case_aux c "step_bracket_end"
  | Case_aux c "step_labelof"
  | Case_aux c "step_getpc"
  | Case_aux c "step_mknav"
  | Case_aux c "step_tosum"
  | Case_aux c "step_loc"
  | Case_aux c "step_locret"
  ].

Require Import Relation_Operators.
Definition steps := clos_refl_trans Cfg step.
Notation "c1 '-->*' c2" := (steps c1 c2) (at level 55, c2 at next level).

Inductive final : Cfg -> Prop :=
  | fa : forall pc r a, 
    final << pc, r, empty, [], CA a >>.
Hint Constructors final.

(** Checking that final_cfgs are really stuck (sanity check)
-- the original unwind rule could be applied forever.
*)
Lemma final_stuck : forall c,
  final c ->
  ~(exists c', c --> c').
Proof.
  intros c H Hc. destruct Hc as [c' Hc].
  (step_cases (destruct Hc) Case); invsc H.
Qed.

Definition binds r x := exists a:Atom, maps r x a.
Definition extending r x (P:Env->Tm->Prop) := (fun tm => forall a:Atom, P ((x,a)::r) tm).

Inductive wf_tm : Env -> Tm -> Prop :=
  (* var *)
  | wf_var : forall r x, binds r x -> wf_tm r (TVar x)
  (* const *)
  | wf_const : forall r c, wf_tm r (TConst c)
  (* let *)
  | wf_let : forall r x t1 t2, wf_tm r t1 -> extending r x wf_tm t2 -> wf_tm r (TLet x t1 t2)
  (* abs *)
  | wf_abs : forall r x t, extending r x wf_tm t -> wf_tm r (TAbs x t)
  (* app *)
  | wf_app : forall r x1 x2, binds r x1 -> binds r x2 -> wf_tm r (TApp x1 x2)
  (* prod *)
  | wf_prod : forall r x1 x2, binds r x1 -> binds r x2 -> wf_tm r (TProd x1 x2)
  (* proj *)
  | wf_proj : forall r d x, binds r x -> wf_tm r (TProj d x)
  (* inx *)
  | wf_inx : forall r d x, binds r x -> wf_tm r (TInx d x)
  (* match *)
  | wf_match : forall r x y t1 t2,
    binds r x -> extending r y wf_tm t1 -> extending r y wf_tm t2 -> wf_tm r (TMatch x y t1 t2)
  (* tag *)
  | wf_tag : forall r x, binds r x -> wf_tm r (TTag x)
  (* bop *)
  | wf_bop : forall r bo x1 x2, binds r x1 -> binds r x2 -> wf_tm r (TBOp bo x1 x2)
  (* brackets *)
  | wf_bracket : forall r x t, binds r x -> wf_tm r t -> wf_tm r (TBracket x t)
  (* label_of *)
  | wf_labelof : forall r x, binds r x -> wf_tm r (TLabelOf x)
  (* get_pc *)
  | wf_getpc : forall r, wf_tm r TGetPc
  (* mk_nav *)
  | wf_mknav : forall r x, binds r x -> wf_tm r (TMkNav x)
  (* to_sum *)
  | wf_tosum : forall r x, binds r x -> wf_tm r (TToSum x)
  (* loc *)
  | wf_loc : forall r loc t, wf_tm r t -> wf_tm r (TLoc loc t).

Inductive wf_atom : Env -> Atom -> Prop :=
  | wf_a : forall r b l, wf_box r b -> wf_atom r (b,l)
with wf_box : Env -> Box -> Prop :=
  | wfb_d : forall r e st pt, wf_atom r st -> wf_atom r pt -> wf_box r (D e st pt)
  | wfv_const : forall r c, wf_box r (V (VConst c))
  | wfv_prod : forall r a1 a2, wf_atom r a1 -> wf_atom r a2 -> wf_box r (V (VProd a1 a2))
  | wfv_inx : forall r d a, wf_atom r a -> wf_box r (V (VInx d a))
  | wfv_clos : forall r r' x t, 
    extending r' x wf_tm t -> wf_env r' -> wf_box r (V (VClos r' x t))
with wf_env : Env -> Prop := 
  | wfe_nil : wf_env []
  | wfe_cons : forall x a r, wf_atom r a -> wf_env r -> wf_env ((x,a)::r).

Inductive wf_trace_cont : Env -> Atom -> Cont -> Prop :=
  | wfk_nil : forall r, wf_trace_cont r empty []
  | wf_rlet : forall r x t k s, 
    extending r x wf_tm t -> wf_trace_cont r s k -> wf_trace_cont r s (RLet x t :: k)
  | wf_rbrk : forall r l pc k s, 
    wf_trace_cont r s k -> wf_trace_cont r s (RBrk l pc :: k)
  | wf_rret : forall r r' k s, wf_trace_cont r' s k -> wf_trace_cont r s (RRet r' :: k)
  | wf_rloc : forall r k loc s, 
    wf_trace_cont r s k -> wf_trace_cont r (push loc bot s) (RLoc :: k).

Inductive wf_config : Cfg -> Prop :=
  | wf_cr : forall pc r s k t, 
    wf_env r -> wf_tm r t -> wf_trace_cont r s k -> wf_config <<pc,r,s,k,CR t>>
  | wf_ca : forall pc r s k a, 
    wf_env r -> wf_atom r a -> wf_trace_cont r s k -> wf_config <<pc,r,s,k,CA a>>.

Hint Constructors wf_tm wf_atom wf_box wf_env wf_trace_cont wf_config.

Ltac binding := 
  repeat match goal with 
           | H : binds _ _ |- _ => 
             unfold binds in H; 
             let b := fresh "b" in
             let l := fresh "l" in
             let Hmap := fresh "Hmap" in
             destruct H as [[b l] Hmap]
         end.

Ltac by_rule Hrule := eexists; eapply Hrule; eauto; simpl; auto.

Theorem progress_imp : forall c,
  wf_config c -> final c \/ exists c', c --> c'.
Proof.
  intros c Hwf. invsc Hwf.
  Case "CR".
    right.
    (Tm_cases (destruct t) SCase); invsc H0; eauto.
    SCase "TVar". binding. by_rule step_var.
    SCase "TApp".
      binding. destruct b0; try destruct v1; try solve [by_rule step_app_error].
      by_rule step_app.
    SCase "TProd". binding. by_rule step_prod.
    SCase "TProj". binding. destruct b; try destruct v0; try solve [by_rule step_proj_error].
      by_rule step_proj.
    SCase "TInx". binding. by_rule step_inx.
    SCase "TMatch". binding. destruct b; try destruct v1; try solve [by_rule step_match_error].
      by_rule step_match.
    SCase "TTag". binding. by_rule step_tag.
    SCase "TBOp". binding. binding. by_rule step_bop.
    SCase "TBracket". binding. 
      destruct b; try destruct v0; try destruct c; try solve [by_rule step_bracket_start_error].
      by_rule step_bracket_start.
    SCase "TLabelOf". binding. by_rule step_labelof.
    SCase "TMkNav". binding. by_rule step_mknav.
    SCase "TToSum". binding. by_rule step_tosum.
  Case "CA".
    destruct k.
    SCase "Done".
      left. invsc H1. auto.
    SCase "Has frame". invsc H1; right.
      SSCase "RLet". by_rule step_let_bind.
      SSCase "RBrk". destruct a. by_rule step_bracket_end.
      SSCase "RRet". by_rule step_ret.
      SSCase "RLoc". destruct a. by_rule step_locret.
Qed.

(** * Proving this small-step semantics equivalent to the big-step one *)

Lemma big_to_small : forall r s t pc a pc' k,
  r; s |= t, pc ==> a, pc' ->
  <<pc, r, s, k, CR t>> -->* <<pc', r, s, k, CA a>>.
Proof.
  intros r s t pc a pc' k He. generalize dependent k.
  (eval_cases (induction He) Case); intro k; try solve [apply rt_step; eauto].
  Case "eval_let".
    eapply rt_trans. auto using rt_step, step_let_start. 
    eapply rt_trans. apply IHHe1.
    eapply rt_trans. auto using rt_step, step_let_bind. 
    eapply rt_trans. apply IHHe2.
    auto using rt_step, step_ret.
  Case "eval_app".
    eapply rt_trans. apply rt_step. eapply step_app; eauto.
    eapply rt_trans. apply IHHe.
    auto using rt_step, step_ret.
  Case "eval_match".
    eapply rt_trans. apply rt_step. eapply step_match; eauto.
    eapply rt_trans. apply IHHe.
    auto using rt_step, step_ret.
  Case "eval_bracket".
    eapply rt_trans. apply rt_step. eapply step_bracket_start; eauto.
    eapply rt_trans. apply IHHe.
    apply  rt_step. eapply step_bracket_end; eauto.
  Case "eval_loc".
    eapply rt_trans. apply rt_step. eapply step_loc; eauto.
    eapply rt_trans. apply IHHe.
    apply rt_step. apply step_locret.
Qed.

Reserved Notation "r ; s |=c cfg , pc1 ==> a , pc2" (at level 80).
Inductive evalCfg : Env -> Atom -> Cfg -> Lab -> Atom -> Lab -> Prop :=
| evalCfgT : forall r s pc r' s' pc' k t a a' pc'' pc''',
    r'; s' |= t, pc' ==> a, pc'' ->
    r; s |=c << pc'', r', s', k, CA a >>, pc ==> a', pc''' ->
    r; s |=c << pc', r', s', k, CR t >>, pc ==> a', pc'''
| evalCfgNilA : forall r s pc pc' a,
    r; s |=c << pc, r, empty, [], CA a >>, pc' ==> a, pc
| evalCfgLetA : forall r s pc r' s' pc' x t k a1 a pcfinal,
    r; s |=c << pc', (x,a1)::r', s', RRet r' :: k, CR t >>, pc ==> a, pcfinal ->
    r; s |=c << pc', r', s', RLet x t :: k, CA a1 >>, pc ==> a, pcfinal
| evalCfgReturnA : forall r s pc r' s' pc' r1 k a1 a pcfinal,
    r; s |=c << pc', r1, s', k, CA a1 >>, pc ==> a, pcfinal ->
    r; s |=c << pc', r', s', RRet r1 :: k, CA a1 >>, pc ==> a, pcfinal
| evalCfgLocA : forall r s pc r' loc pc'' s' pc' k b l a pcfinal,
    r; s |=c << pc', r', s', k, CA (trace_propagation b loc pc',l) >>, pc ==> a, pcfinal ->
    r; s |=c << pc', r', push loc pc'' s', RLoc :: k, CA (b,l) >>, pc ==> a, pcfinal
| evalCfgBracketA : forall r s pc r' s' pc' l' pc'' k b l'' a pcfinal,
    r; s |=c << pc'', r', s', k, 
            CA (if flows_dec (l''\_/pc') (l'\_/pc'') then b else D eBracket s' empty,l') >>, pc ==> 
         a, pcfinal ->
    r; s |=c << pc', r', s', RBrk l' pc'' :: k, CA (b,l'') >>, pc ==> a, pcfinal
where "r ; s |=c cfg , pc1 ==> a , pc2" := (evalCfg r s cfg pc1 a pc2).
Hint Constructors evalCfg.

(** [evalCfg] is preserved by stepping back. *)
Lemma evalCfg_anti_step : forall r s pc pc' a conf conf',
  r;s |=c conf', pc ==> a, pc' ->
  conf --> conf' ->
  r;s |=c conf, pc ==> a, pc'.
Proof.
intros r s pc pc' a conf conf' HevalCfg Hstep.
step_cases (invsc Hstep) Case; try solve [invs HevalCfg; eauto | eapply evalCfgT; eauto; auto].
Case "step_let_start". invs HevalCfg. invs H10. invs H13. invs H12. eauto.
Case "step_app". invs HevalCfg. invs H12. eauto.
Case "step_match". invs HevalCfg. invs H11. eauto.
Case "step_bracket_start". 
  invs HevalCfg. invs H11. eapply evalCfgT; eauto; auto.
Case "step_labelof". eapply evalCfgT; eauto. eauto using eval_label_of.
Case "step_loc". invs HevalCfg. invs H10. eapply evalCfgT; eauto. eapply eval_loc; eauto.
Case "step_locret". apply evalCfgLocA. assumption.
Qed.

(** Generalization of the previous lemma to an arbitrary
    number of steps. *)
Lemma evalCfg_anti_step_star : forall r s pc pc' a conf conf',
  r;s |=c conf', pc ==> a, pc' ->
  conf -->* conf' ->
  r;s |=c conf, pc ==> a, pc'.
Proof.
intros r s pc pc' a conf conf' HevalCfg Hstep.
induction Hstep; eauto using evalCfg_anti_step.
Qed.

(** The second part of adequacy. *)
Lemma small_to_big : forall r pc r' pc' t a,
  << pc, r, empty, [], CR t >> -->* << pc', r', empty, [], CA a >> ->
  r; empty |= t, pc ==> a, pc' /\ r' = r.
Proof.
intros r pc r' pc' t a Hstep.
assert (r';empty |=c << pc, r, empty, [], CR t >>, pc ==> a, pc').
 eapply evalCfg_anti_step_star; eauto.
invs H. invs H11. auto.
Qed.

(** Finally, we prove adequacy. *)
Theorem adequacy: forall r pc pc' t a,
  r; empty |= t, pc ==> a, pc' <-> 
  << pc, r, empty, [], CR t >> -->* << pc', r, empty, [], CA a >>.
Proof.
intros r pc pc' t a. split; intro H.
auto using big_to_small.
apply small_to_big in H as [Heval _]. assumption.
Qed.
